This commit is contained in:
Roger Wang
2026-03-25 02:47:31 -07:00
parent 2e67fa756d
commit 7cd8824477
6 changed files with 1120 additions and 2 deletions
+454
View File
@@ -0,0 +1,454 @@
"""
Benchmark: SM103 (B300) FP4 Ultra GEMM vs SM100 (B200) NVFP4 GEMM
===================================================================
This benchmark compares the performance of the SM103-optimized FP4 Ultra
GEMM kernel against the default SM100 NVFP4 GEMM kernel, both running on
B300 hardware.
SM103 kernels use:
- K=768 tile (vs K=256 on SM100)
- FP4 Ultra MMA (UltraVs16) schedule
- NoSmemWarpSpecialized epilogue
- Sm103BlockScaledConfig scale factor layout
Usage:
python benchmarks/kernels/benchmark_nvfp4_sm103.py [--mode gemm|quant|e2e|all]
Requirements:
- B300 GPU (SM103 / compute capability 10.3)
- CUDA >= 12.9
- vLLM built with ENABLE_NVFP4_SM100=1 and SM103 support
"""
import argparse
import time
from typing import Optional
import torch
# ============================================================================
# Helpers
# ============================================================================
def round_up(x: int, y: int) -> int:
return ((x + y - 1) // y) * y
def get_sm_version() -> int:
"""Return SM version as integer (e.g., 100, 103, 120)."""
cap = torch.cuda.get_device_capability()
return cap[0] * 10 + cap[1]
def create_nvfp4_tensors(
m: int, n: int, k: int, dtype: torch.dtype = torch.bfloat16
) -> dict:
"""
Create synthetic NVFP4 GEMM input tensors (A, B, scales, alpha).
A: [m, k/2] uint8 (packed FP4)
B: [n, k/2] uint8 (packed FP4, column-major)
A_sf: [round_up(m,128), round_up(k/16,4)] float8_e4m3fn (swizzled)
B_sf: [round_up(n,128), round_up(k/16,4)] float8_e4m3fn (swizzled)
alpha: [1] float32
D: [m, n] output
"""
# Packed FP4 data (random bytes -- content doesn't affect timing)
A = torch.randint(0, 256, (m, k // 2), dtype=torch.uint8, device="cuda")
B = torch.randint(0, 256, (n, k // 2), dtype=torch.uint8, device="cuda")
# Scale factors (padded, will be swizzled separately for SM100/SM103)
sf_m = round_up(m, 128)
sf_n = round_up(n, 128)
sf_k = round_up(k // 16, 4)
# Create as int32 (raw bytes, same shape as expected by CUTLASS)
A_sf = torch.randint(
0, 256, (sf_m, sf_k), dtype=torch.uint8, device="cuda"
).view(torch.float8_e4m3fn)
B_sf = torch.randint(
0, 256, (sf_n, sf_k), dtype=torch.uint8, device="cuda"
).view(torch.float8_e4m3fn)
# Global alpha
alpha = torch.tensor([1.0], dtype=torch.float32, device="cuda")
# Output
D = torch.empty(m, n, dtype=dtype, device="cuda")
return {
"A": A,
"B": B,
"A_sf": A_sf,
"B_sf": B_sf,
"alpha": alpha,
"D": D,
}
def create_quant_tensors(
m: int, n: int, dtype: torch.dtype = torch.bfloat16
) -> dict:
"""Create inputs for activation quantization benchmark."""
input_tensor = torch.randn(m, n, dtype=dtype, device="cuda")
global_scale = torch.tensor([0.5], dtype=torch.float32, device="cuda")
return {"input": input_tensor, "global_scale": global_scale}
def bench_fn(
fn,
warmup: int = 20,
iters: int = 100,
sync: bool = True,
) -> float:
"""Benchmark a function, returning median time in microseconds."""
# Warmup
for _ in range(warmup):
fn()
if sync:
torch.cuda.synchronize()
# Timed iterations using CUDA events
start_events = [torch.cuda.Event(enable_timing=True) for _ in range(iters)]
end_events = [torch.cuda.Event(enable_timing=True) for _ in range(iters)]
for i in range(iters):
start_events[i].record()
fn()
end_events[i].record()
torch.cuda.synchronize()
times = [s.elapsed_time(e) * 1000 for s, e in zip(start_events, end_events)]
times.sort()
# Return median in microseconds
return times[len(times) // 2]
# ============================================================================
# GEMM Benchmark
# ============================================================================
def benchmark_gemm(
m_sizes: list[int],
n: int = 7168,
k: int = 7168,
dtype: torch.dtype = torch.bfloat16,
) -> list[dict]:
"""
Benchmark SM100 vs SM103 NVFP4 GEMM kernels.
Since we can't call internal CUTLASS kernels directly from Python,
this benchmark uses the top-level cutlass_scaled_fp4_mm dispatch.
On B300, the dispatcher routes to sm103a; we also time the sm100a
path by calling it directly if available.
"""
try:
from vllm._C import ops as vllm_ops # type: ignore
except ImportError:
print("ERROR: vLLM C extensions not built. Build with: pip install -e .")
return []
results = []
sm = get_sm_version()
for m in m_sizes:
tensors = create_nvfp4_tensors(m, n, k, dtype)
D, A, B = tensors["D"], tensors["A"], tensors["B"]
A_sf, B_sf, alpha = tensors["A_sf"], tensors["B_sf"], tensors["alpha"]
# --- SM100 kernel (baseline on B300, runs via forward compatibility) ---
# We create SM100-layout scale factors for the SM100 kernel.
# The top-level dispatch on SM103 calls sm103a, so for SM100 baseline
# we'd need to call cutlass_scaled_fp4_mm_sm100a directly.
# Since that's not directly exposed, we measure the default dispatch
# and note which path it takes.
def run_default():
vllm_ops.cutlass_scaled_fp4_mm(D, A, B, A_sf, B_sf, alpha)
time_us = bench_fn(run_default, warmup=20, iters=100)
# Compute effective TFLOPS
# FP4 GEMM: 2*M*N*K FLOPs (multiply-add)
flops = 2.0 * m * n * k
tflops = flops / (time_us * 1e-6) / 1e12
kernel_name = f"SM{sm} (default dispatch)"
results.append({
"M": m,
"N": n,
"K": k,
"kernel": kernel_name,
"time_us": time_us,
"tflops": tflops,
})
return results
# ============================================================================
# Quantization Benchmark
# ============================================================================
def benchmark_quant(
m_sizes: list[int],
n: int = 7168,
dtype: torch.dtype = torch.bfloat16,
) -> list[dict]:
"""
Benchmark SM100 vs SM103 activation quantization (BF16 -> NVFP4).
"""
try:
from vllm._C import ops as vllm_ops # type: ignore
except ImportError:
print("ERROR: vLLM C extensions not built.")
return []
results = []
for m in m_sizes:
tensors = create_quant_tensors(m, n, dtype)
input_t = tensors["input"]
global_scale = tensors["global_scale"]
# SM100 quantization (swizzled layout)
def run_sm100_quant():
vllm_ops.scaled_fp4_quant(input_t, global_scale, True)
time_sm100 = bench_fn(run_sm100_quant, warmup=20, iters=100)
results.append({
"M": m,
"N": n,
"kernel": "SM100 quant (swizzled)",
"time_us": time_sm100,
"throughput_gb_s": (m * n * 2) / (time_sm100 * 1e-6) / 1e9,
})
# SM103 quantization would use scaled_fp4_quant_sm103a
# (requires the new op to be registered; placeholder for when available)
return results
# ============================================================================
# SF Layout Conversion Benchmark
# ============================================================================
def benchmark_sf_conversion(
m_sizes: list[int],
k: int = 7168,
) -> list[dict]:
"""
Benchmark the SM100 <-> SM103 scale factor layout conversion kernel.
This measures the overhead of converting scale factors between layouts,
which happens once at model load time for weights.
"""
try:
from vllm._C import ops as vllm_ops # type: ignore
except ImportError:
print("ERROR: vLLM C extensions not built.")
return []
results = []
for m in m_sizes:
sf_m = round_up(m, 128)
sf_k = round_up(k // 16, 4)
# Create source SF tensor (SM100 layout)
src = torch.randint(
0, 256, (sf_m, sf_k), dtype=torch.uint8, device="cuda"
).view(torch.float8_e4m3fn)
# Allocate destination (same shape)
dst = torch.empty_like(src)
# Benchmark SM100 -> SM103 conversion
def run_convert():
vllm_ops.convert_sf_layout_sm100_to_sm103(dst, src)
time_us = bench_fn(run_convert, warmup=20, iters=200)
results.append({
"M": m,
"K": k,
"sf_shape": f"{sf_m}x{sf_k}",
"kernel": "SM100->SM103 SF convert",
"time_us": time_us,
"throughput_gb_s": (sf_m * sf_k) / (time_us * 1e-6) / 1e9,
})
return results
# ============================================================================
# End-to-End Benchmark (Quant + GEMM)
# ============================================================================
def benchmark_e2e(
m_sizes: list[int],
n: int = 7168,
k: int = 7168,
dtype: torch.dtype = torch.bfloat16,
) -> list[dict]:
"""
Benchmark the full NVFP4 inference path: quantize activations + GEMM.
This measures what a real transformer linear layer does:
1. Quantize BF16 activations to NVFP4 (with block scales)
2. NVFP4 x NVFP4 GEMM
"""
try:
from vllm._C import ops as vllm_ops # type: ignore
except ImportError:
print("ERROR: vLLM C extensions not built.")
return []
results = []
sm = get_sm_version()
for m in m_sizes:
# Create activation input
activation = torch.randn(m, k, dtype=dtype, device="cuda")
global_scale = torch.tensor([0.5], dtype=torch.float32, device="cuda")
# Create weight (pre-quantized, SM100 layout for default)
B = torch.randint(0, 256, (n, k // 2), dtype=torch.uint8, device="cuda")
sf_n = round_up(n, 128)
sf_k = round_up(k // 16, 4)
B_sf = torch.randint(
0, 256, (sf_n, sf_k), dtype=torch.uint8, device="cuda"
).view(torch.float8_e4m3fn)
alpha = torch.tensor([1.0], dtype=torch.float32, device="cuda")
D = torch.empty(m, n, dtype=dtype, device="cuda")
# Pre-allocate quant output
A_packed = torch.empty(m, k // 2, dtype=torch.uint8, device="cuda")
def run_e2e():
# Step 1: Quantize activations
A_q, A_sf = vllm_ops.scaled_fp4_quant(
activation, global_scale, True
)
# Step 2: GEMM
vllm_ops.cutlass_scaled_fp4_mm(D, A_q, B, A_sf, B_sf, alpha)
time_us = bench_fn(run_e2e, warmup=10, iters=50)
flops = 2.0 * m * n * k
tflops = flops / (time_us * 1e-6) / 1e12
results.append({
"M": m,
"N": n,
"K": k,
"kernel": f"SM{sm} E2E (quant+GEMM)",
"time_us": time_us,
"tflops": tflops,
})
return results
# ============================================================================
# Main
# ============================================================================
def print_results(results: list[dict], title: str):
if not results:
return
print(f"\n{'=' * 80}")
print(f" {title}")
print(f"{'=' * 80}")
# Determine columns from first result
cols = list(results[0].keys())
# Header
header = " | ".join(f"{c:>15s}" for c in cols)
print(header)
print("-" * len(header))
for r in results:
row = []
for c in cols:
v = r[c]
if isinstance(v, float):
row.append(f"{v:>15.2f}")
elif isinstance(v, int):
row.append(f"{v:>15d}")
else:
row.append(f"{v:>15s}")
print(" | ".join(row))
def main():
parser = argparse.ArgumentParser(
description="Benchmark NVFP4 SM103 vs SM100 kernels"
)
parser.add_argument(
"--mode",
choices=["gemm", "quant", "sf_convert", "e2e", "all"],
default="all",
help="Which benchmark to run",
)
parser.add_argument(
"--n", type=int, default=7168, help="N dimension (default: 7168, DeepSeek)"
)
parser.add_argument(
"--k", type=int, default=7168, help="K dimension (default: 7168, DeepSeek)"
)
args = parser.parse_args()
sm = get_sm_version()
print(f"GPU: {torch.cuda.get_device_name()}")
print(f"SM version: {sm}")
print(f"CUDA version: {torch.version.cuda}")
if sm < 100:
print("ERROR: This benchmark requires SM100+ (Blackwell) GPU.")
return
if sm == 103:
print("NOTE: Running on SM103 (B300) -- SM103 kernels will be used.")
else:
print(f"NOTE: Running on SM{sm} -- SM100 kernels will be used.")
# Problem sizes typical for LLM inference
# Small M = decode, large M = prefill
m_sizes = [1, 4, 8, 16, 32, 64, 128, 256, 512, 1024, 2048, 4096]
if args.mode in ("gemm", "all"):
results = benchmark_gemm(m_sizes, n=args.n, k=args.k)
print_results(results, f"NVFP4 GEMM Benchmark (N={args.n}, K={args.k})")
if args.mode in ("quant", "all"):
results = benchmark_quant(m_sizes, n=args.k)
print_results(results, f"NVFP4 Activation Quantization (N={args.k})")
if args.mode in ("sf_convert", "all"):
# Use N dimension for SF conversion (weight matrix rows)
sf_m_sizes = [1024, 2048, 4096, 7168, 8192, 14336, 16384]
results = benchmark_sf_conversion(sf_m_sizes, k=args.k)
print_results(results, "SF Layout Conversion SM100 <-> SM103")
if args.mode in ("e2e", "all"):
results = benchmark_e2e(m_sizes, n=args.n, k=args.k)
print_results(
results,
f"End-to-End NVFP4 (Quant+GEMM, N={args.n}, K={args.k})",
)
if __name__ == "__main__":
main()
@@ -171,8 +171,254 @@ __global__ void __launch_bounds__(512, VLLM_BLOCKS_PER_SM(512))
}
}
// ============================================================================
// SM103 (B300) activation quantization kernel.
//
// Identical to the SM100 cvt_fp16_to_fp4 except it writes scale factors
// in the SM103 swizzled layout (Sm103BlockScaledConfig).
// ============================================================================
template <class Type, bool UE8M0_SF = false>
__global__ void __launch_bounds__(512, VLLM_BLOCKS_PER_SM(512))
cvt_fp16_to_fp4_sm103(int32_t numRows, int32_t numCols,
int32_t num_padded_cols,
Type const* __restrict__ in,
float const* __restrict__ SFScale,
uint32_t* __restrict__ out,
uint32_t* __restrict__ SFout) {
using PackedVec = vllm::PackedVec<Type, CVT_FP4_PACK16>;
static constexpr int CVT_FP4_NUM_THREADS_PER_SF =
(CVT_FP4_SF_VEC_SIZE / CVT_FP4_ELTS_PER_THREAD);
static_assert(sizeof(PackedVec) == sizeof(Type) * CVT_FP4_ELTS_PER_THREAD,
"Vec size is not matched.");
int32_t const numKTiles = (numCols + 63) / 64;
int sf_m = round_up<int>(numRows, 128);
int32_t const colIdx = blockDim.x * blockIdx.y + threadIdx.x;
int elem_idx = colIdx * CVT_FP4_ELTS_PER_THREAD;
float const global_scale = (SFScale == nullptr) ? 1.0f : SFScale[0];
for (int rowIdx = blockIdx.x; rowIdx < sf_m; rowIdx += gridDim.x) {
if (colIdx < num_padded_cols) {
PackedVec in_vec;
int64_t inOffset = rowIdx * (numCols / CVT_FP4_ELTS_PER_THREAD) + colIdx;
bool valid = (rowIdx < numRows) && (elem_idx < numCols);
if constexpr (CVT_FP4_PACK16) {
ld256_cg_or_zero(reinterpret_cast<u32x8_t&>(in_vec),
&reinterpret_cast<const uint32_t*>(in)[inOffset * 8],
valid);
} else {
ld128_cg_or_zero(reinterpret_cast<uint4&>(in_vec),
&reinterpret_cast<const uint32_t*>(in)[inOffset * 4],
valid);
}
// SM103: Use SM103-specific SF offset function
auto sf_out =
cvt_quant_to_fp4_get_sf_out_offset_sm103<uint32_t,
CVT_FP4_NUM_THREADS_PER_SF>(
rowIdx, colIdx, numKTiles, SFout);
auto out_val =
cvt_warp_fp16_to_fp4<Type, CVT_FP4_NUM_THREADS_PER_SF, UE8M0_SF>(
in_vec, global_scale, sf_out);
if (valid) {
if constexpr (CVT_FP4_PACK16) {
int64_t outOffset = rowIdx * (numCols / 8) + colIdx * 2;
uint64_t packed64 =
(uint64_t(out_val.hi) << 32) | uint64_t(out_val.lo);
reinterpret_cast<uint64_t*>(out)[outOffset >> 1] = packed64;
} else {
out[inOffset] = out_val;
}
}
}
}
}
// ============================================================================
// Scale factor layout conversion: SM100 <-> SM103
//
// Converts an already-swizzled SF tensor between SM100 and SM103 layouts.
// Both layouts use the same 512-byte tile structure (128 M-rows x 4 K-cols)
// but arrange bytes differently within each tile.
//
// SM100 offset: outerM(=mIdx%32)*16 + innerM(=(mIdx/32)%4)*4 + innerK
// SM103 offset: m8(=(mIdx/16)%8)*16 + m4a(=(mIdx/4)%4)*128 + m4b(=mIdx%4)*4
// + innerK
// ============================================================================
__global__ void convert_sf_sm100_to_sm103_kernel(
const uint8_t* __restrict__ src,
uint8_t* __restrict__ dst,
int32_t numMTiles,
int32_t numKTiles) {
// Each thread converts one byte (one SF value).
// Grid: numMTiles * numKTiles blocks, 512 threads per block.
int32_t tile_idx = blockIdx.x;
int32_t mTileIdx = tile_idx / numKTiles;
int32_t kTileIdx = tile_idx % numKTiles;
// Each tile is 512 bytes: 128 M-positions x 4 K-positions.
int32_t local_idx = threadIdx.x; // 0..511
if (mTileIdx >= numMTiles) return;
int64_t tile_base = static_cast<int64_t>(tile_idx) << 9;
// Decode this thread's (mLocal, kLocal) from a simple linear index.
int32_t mLocal = local_idx >> 2; // 0..127
int32_t kLocal = local_idx & 3; // 0..3
// Compute SM100 source offset within tile.
int32_t outerMIdx = mLocal & 31;
int32_t innerMIdx = (mLocal >> 5) & 3;
int32_t sm100_off = (outerMIdx << 4) | (innerMIdx << 2) | kLocal;
// Compute SM103 destination offset within tile.
int32_t m4b = mLocal & 3;
int32_t m4a = (mLocal >> 2) & 3;
int32_t m8 = (mLocal >> 4) & 7;
int32_t sm103_off = (m8 << 4) | (m4a << 7) | (m4b << 2) | kLocal;
dst[tile_base + sm103_off] = src[tile_base + sm100_off];
}
__global__ void convert_sf_sm103_to_sm100_kernel(
const uint8_t* __restrict__ src,
uint8_t* __restrict__ dst,
int32_t numMTiles,
int32_t numKTiles) {
int32_t tile_idx = blockIdx.x;
int32_t mTileIdx = tile_idx / numKTiles;
if (mTileIdx >= numMTiles) return;
int32_t local_idx = threadIdx.x;
int64_t tile_base = static_cast<int64_t>(tile_idx) << 9;
int32_t mLocal = local_idx >> 2;
int32_t kLocal = local_idx & 3;
// SM103 source offset
int32_t m4b = mLocal & 3;
int32_t m4a = (mLocal >> 2) & 3;
int32_t m8 = (mLocal >> 4) & 7;
int32_t sm103_off = (m8 << 4) | (m4a << 7) | (m4b << 2) | kLocal;
// SM100 destination offset
int32_t outerMIdx = mLocal & 31;
int32_t innerMIdx = (mLocal >> 5) & 3;
int32_t sm100_off = (outerMIdx << 4) | (innerMIdx << 2) | kLocal;
dst[tile_base + sm100_off] = src[tile_base + sm103_off];
}
} // namespace vllm
// ============================================================================
// Host entry: SM103 activation quantization
// ============================================================================
void scaled_fp4_quant_sm103a(torch::Tensor const& output,
torch::Tensor const& input,
torch::Tensor const& output_sf,
torch::Tensor const& input_sf) {
int32_t m = input.size(0);
int32_t n = input.size(1);
TORCH_CHECK(n % 16 == 0, "The N dimension must be multiple of 16.");
TORCH_CHECK(input.scalar_type() == at::ScalarType::Half ||
input.scalar_type() == at::ScalarType::BFloat16,
"Unsupported input data type for quantize_to_fp4.");
int multiProcessorCount =
get_device_attribute(cudaDevAttrMultiProcessorCount, -1);
auto input_sf_ptr = static_cast<float const*>(input_sf.data_ptr());
auto sf_out = static_cast<int32_t*>(output_sf.data_ptr());
auto output_ptr = static_cast<int64_t*>(output.data_ptr());
const at::cuda::OptionalCUDAGuard device_guard(device_of(input));
auto stream = at::cuda::getCurrentCUDAStream(input.get_device());
int sf_n_unpadded = int(n / CVT_FP4_SF_VEC_SIZE);
dim3 block(std::min(int(n / ELTS_PER_THREAD), 512));
int const numBlocksPerSM =
vllm_runtime_blocks_per_sm(static_cast<int>(block.x));
// SM103 always uses swizzled layout (the SM103 variant)
int sf_n_int = int(vllm::round_up(sf_n_unpadded, 4) / 4);
int32_t num_padded_cols =
sf_n_int * 4 * CVT_FP4_SF_VEC_SIZE / CVT_FP4_ELTS_PER_THREAD;
int grid_y = vllm::div_round_up(num_padded_cols, static_cast<int>(block.x));
int grid_x =
std::min(vllm::computeEffectiveRows(m),
std::max(1, (multiProcessorCount * numBlocksPerSM) / grid_y));
dim3 grid(grid_x, grid_y);
VLLM_DISPATCH_HALF_TYPES(input.scalar_type(), "nvfp4_quant_sm103", [&] {
using cuda_type = vllm::CUDATypeConverter<scalar_t>::Type;
auto input_ptr = static_cast<cuda_type const*>(input.data_ptr());
vllm::cvt_fp16_to_fp4_sm103<cuda_type, false><<<grid, block, 0, stream>>>(
m, n, num_padded_cols, input_ptr, input_sf_ptr,
reinterpret_cast<uint32_t*>(output_ptr),
reinterpret_cast<uint32_t*>(sf_out));
});
}
// ============================================================================
// Host entry: SF layout conversion SM100 <-> SM103
// ============================================================================
void convert_sf_layout_sm100_to_sm103(torch::Tensor& dst,
torch::Tensor const& src) {
TORCH_CHECK(src.is_contiguous(), "Source SF tensor must be contiguous");
TORCH_CHECK(dst.is_contiguous(), "Destination SF tensor must be contiguous");
TORCH_CHECK(src.numel() == dst.numel(),
"Source and destination must have the same number of elements");
// SF tensors are stored as int32 with shape (rounded_m, rounded_k / 4)
// Total bytes = rounded_m * (rounded_k / 4) * 4 = rounded_m * rounded_k
int64_t total_bytes = src.numel() * src.element_size();
int32_t numMTiles = src.size(0) / 128;
int32_t numKTiles = total_bytes / (numMTiles * 512);
const at::cuda::OptionalCUDAGuard device_guard(device_of(src));
auto stream = at::cuda::getCurrentCUDAStream(src.get_device());
int32_t num_tiles = numMTiles * numKTiles;
dim3 grid(num_tiles);
dim3 block(512);
vllm::convert_sf_sm100_to_sm103_kernel<<<grid, block, 0, stream>>>(
static_cast<const uint8_t*>(src.data_ptr()),
static_cast<uint8_t*>(dst.data_ptr()),
numMTiles, numKTiles);
}
void convert_sf_layout_sm103_to_sm100(torch::Tensor& dst,
torch::Tensor const& src) {
TORCH_CHECK(src.is_contiguous() && dst.is_contiguous());
TORCH_CHECK(src.numel() == dst.numel());
int64_t total_bytes = src.numel() * src.element_size();
int32_t numMTiles = src.size(0) / 128;
int32_t numKTiles = total_bytes / (numMTiles * 512);
const at::cuda::OptionalCUDAGuard device_guard(device_of(src));
auto stream = at::cuda::getCurrentCUDAStream(src.get_device());
int32_t num_tiles = numMTiles * numKTiles;
vllm::convert_sf_sm103_to_sm100_kernel<<<dim3(num_tiles), dim3(512), 0, stream>>>(
static_cast<const uint8_t*>(src.data_ptr()),
static_cast<uint8_t*>(dst.data_ptr()),
numMTiles, numKTiles);
}
// ============================================================================
// Original SM100 host entry
// ============================================================================
void scaled_fp4_quant_sm1xxa(torch::Tensor const& output,
torch::Tensor const& input,
torch::Tensor const& output_sf,
@@ -24,6 +24,13 @@ void cutlass_scaled_fp4_mm_sm100a(torch::Tensor& D, torch::Tensor const& A,
torch::Tensor const& A_sf,
torch::Tensor const& B_sf,
torch::Tensor const& alpha);
// SM103 (B300) uses FP4 Ultra MMA -- separate entry point compiled from
// the same source file, guarded by CUTLASS_ARCH_MMA_SM103_SUPPORTED.
void cutlass_scaled_fp4_mm_sm103a(torch::Tensor& D, torch::Tensor const& A,
torch::Tensor const& B,
torch::Tensor const& A_sf,
torch::Tensor const& B_sf,
torch::Tensor const& alpha);
#endif
#if defined ENABLE_NVFP4_SM120 && ENABLE_NVFP4_SM120
@@ -43,6 +50,14 @@ void cutlass_scaled_fp4_mm(torch::Tensor& D, const torch::Tensor& A,
const int32_t sm = get_sm_version_num();
#if defined(ENABLE_NVFP4_SM100) && ENABLE_NVFP4_SM100
// SM103 (B300): Use FP4 Ultra kernels with K=768 tiles for higher
// throughput. Falls through to SM100 path if SM103 kernels werent compiled
// (e.g., CUDA < 12.9).
if (sm == 103) {
cutlass_scaled_fp4_mm_sm103a(D, A, B, A_sf, B_sf, alpha);
return;
}
if (sm >= 100 && sm < 120) {
cutlass_scaled_fp4_mm_sm100a(D, A, B, A_sf, B_sf, alpha);
return;
@@ -36,6 +36,10 @@ using namespace cute;
#if defined(CUTLASS_ARCH_MMA_SM100_SUPPORTED)
// ============================================================================
// SM100 (B200) Tile Configurations
// ============================================================================
// Configuration for M in (256, inf)
struct sm100_fp4_config_default {
using KernelSchedule = cutlass::gemm::collective::KernelScheduleAuto;
@@ -63,6 +67,51 @@ struct sm100_fp4_config_M16 {
using PerSmTileShape_MNK = Shape<_128, _128, _256>;
};
// ============================================================================
// SM103 (B300 / Blackwell Ultra) Tile Configurations
//
// Key differences from SM100:
// - Tile K = 768 is MANDATORY (CUTLASS static_assert)
// - Uses FP4 Ultra MMA instructions (UltraVs16) for higher throughput
// - Uses NoSmem epilogue (saves shared memory for mainloop)
// - 1SM for small M, 2SM for large M (cooperative SM pairs)
// ============================================================================
#if defined(CUTLASS_ARCH_MMA_SM103_SUPPORTED)
// SM103 configuration for M in (256, inf) -- 2SM cooperative execution
struct sm103_fp4_config_default {
// 2SM schedule: two SMs cooperate on one tile for higher throughput
using KernelSchedule = cutlass::gemm::collective::
KernelTmaWarpSpecialized2SmBlockScaledMxNvf4UltraVs16Sm103;
using EpilogueSchedule = cutlass::epilogue::NoSmemWarpSpecialized2Sm;
using TileShape = Shape<_128, _256, Int<768>>;
using ClusterShape = Shape<_2, _2, _1>;
using PerSmTileShape_MNK = Shape<_128, _256, Int<768>>;
};
// SM103 configuration for M in (16, 256] -- 2SM with smaller N tile
struct sm103_fp4_config_M256 {
using KernelSchedule = cutlass::gemm::collective::
KernelTmaWarpSpecialized2SmBlockScaledMxNvf4UltraVs16Sm103;
using EpilogueSchedule = cutlass::epilogue::NoSmemWarpSpecialized2Sm;
using TileShape = Shape<_128, _128, Int<768>>;
using ClusterShape = Shape<_1, _2, _1>;
using PerSmTileShape_MNK = Shape<_128, _128, Int<768>>;
};
// SM103 configuration for M in [1, 16] -- 1SM (decode / small batch)
struct sm103_fp4_config_M16 {
// 1SM schedule: single SM per tile, lower latency for small problems
using KernelSchedule = cutlass::gemm::collective::
KernelTmaWarpSpecialized1SmBlockScaledMxNvf4UltraVs16Sm103;
using EpilogueSchedule = cutlass::epilogue::NoSmemWarpSpecialized1Sm;
using TileShape = Shape<_128, _128, Int<768>>;
using ClusterShape = Shape<_1, _1, _1>;
using PerSmTileShape_MNK = Shape<_128, _128, Int<768>>;
};
#endif // CUTLASS_ARCH_MMA_SM103_SUPPORTED
template <typename Config, typename OutType>
struct Fp4GemmSm100 {
// A matrix configuration
@@ -125,6 +174,99 @@ struct Fp4GemmSm100 {
using LayoutD = decltype(cute::make_layout(make_shape(0, 0, 0), StrideD{}));
};
// ============================================================================
// SM103 GEMM Definition (FP4 Ultra)
//
// SM103 differs from SM100 in several fundamental ways:
// 1. Uses cutlass::arch::Sm103 (separate CollectiveBuilder specialization)
// 2. Element types passed as cute::tuple<DataType, ScaleFactorType>
// (SM100 uses nv_float4_t<float_e2m1_t> wrapper instead)
// 3. Tile K = 768 (SM100 uses K = 256)
// 4. Epilogue uses NoSmemWarpSpecialized (SM100 uses TmaWarpSpecialized)
// 5. Scale factor memory layout uses Sm103BlockScaledConfig
// (different swizzle pattern from SM100's Sm1xxBlockScaledConfig)
//
// IMPORTANT: Scale factor layout compatibility
// SM103 and SM100 use DIFFERENT physical scale factor layouts in memory.
// The activation quantization kernel (scaled_fp4_quant) and the weight
// scale factors in NVFP4 checkpoints must produce/store data in the
// SM103-expected layout when using these kernels. Passing SM100-format
// scale factors to SM103 kernels will produce incorrect results.
// See Sm103BlockScaledConfig::tile_atom_to_shape_SFA for the expected
// layout.
// ============================================================================
#if defined(CUTLASS_ARCH_MMA_SM103_SUPPORTED)
template <typename Config, typename OutType>
struct Fp4GemmSm103 {
// A matrix configuration -- bare float_e2m1_t (not nv_float4_t wrapper)
using ElementA = cutlass::float_e2m1_t;
using ElementSFA = cutlass::float_ue4m3_t;
using LayoutATag = cutlass::layout::RowMajor;
static constexpr int AlignmentA = 32;
// B matrix configuration
using ElementB = cutlass::float_e2m1_t;
using ElementSFB = cutlass::float_ue4m3_t;
using LayoutBTag = cutlass::layout::ColumnMajor;
static constexpr int AlignmentB = 32;
// C/D matrix configuration
using ElementD = OutType;
using ElementC = OutType;
using LayoutCTag = cutlass::layout::RowMajor;
using LayoutDTag = cutlass::layout::RowMajor;
static constexpr int AlignmentD = 128 / cutlass::sizeof_bits<ElementD>::value;
static constexpr int AlignmentC = 128 / cutlass::sizeof_bits<ElementC>::value;
// Kernel functional config
using ElementAccumulator = float;
using ArchTag = cutlass::arch::Sm103;
using OperatorClass = cutlass::arch::OpClassBlockScaledTensorOp;
// Use config's tile shapes (K=768 mandatory for SM103)
using MmaTileShape = typename Config::TileShape;
using ClusterShape = typename Config::ClusterShape;
using PerSmTileShape_MNK = typename Config::PerSmTileShape_MNK;
// Epilogue: SM103 uses NoSmem variant with OpClassTensorOp
// Note: epilogue builder uses Sm100 arch tag (shared epilogue HW)
using CollectiveEpilogue =
typename cutlass::epilogue::collective::CollectiveBuilder<
cutlass::arch::Sm100, cutlass::arch::OpClassTensorOp,
PerSmTileShape_MNK, ClusterShape,
cutlass::epilogue::collective::EpilogueTileAuto, ElementAccumulator,
ElementAccumulator, ElementC, LayoutCTag, AlignmentC, ElementD,
LayoutDTag, AlignmentD,
typename Config::EpilogueSchedule>::CollectiveOp;
// Mainloop: SM103 passes element+SF types as tuples to the builder
using CollectiveMainloop =
typename cutlass::gemm::collective::CollectiveBuilder<
ArchTag, OperatorClass, cute::tuple<ElementA, ElementSFA>, LayoutATag,
AlignmentA, cute::tuple<ElementB, ElementSFB>, LayoutBTag, AlignmentB,
ElementAccumulator, MmaTileShape, ClusterShape,
cutlass::gemm::collective::StageCountAutoCarveout<static_cast<int>(
sizeof(typename CollectiveEpilogue::SharedStorage))>,
typename Config::KernelSchedule>::CollectiveOp;
using GemmKernel = cutlass::gemm::kernel::GemmUniversal<
Shape<int, int, int, int>, CollectiveMainloop, CollectiveEpilogue, void>;
using Gemm = cutlass::gemm::device::GemmUniversalAdapter<GemmKernel>;
using StrideA = typename Gemm::GemmKernel::StrideA;
using LayoutA = decltype(cute::make_layout(make_shape(0, 0, 0), StrideA{}));
using LayoutSFA = typename Gemm::GemmKernel::CollectiveMainloop::LayoutSFA;
using StrideB = typename Gemm::GemmKernel::StrideB;
using LayoutB = decltype(cute::make_layout(make_shape(0, 0, 0), StrideB{}));
using LayoutSFB = typename Gemm::GemmKernel::CollectiveMainloop::LayoutSFB;
using StrideC = typename Gemm::GemmKernel::StrideC;
using LayoutC = decltype(cute::make_layout(make_shape(0, 0, 0), StrideC{}));
using StrideD = typename Gemm::GemmKernel::StrideD;
using LayoutD = decltype(cute::make_layout(make_shape(0, 0, 0), StrideD{}));
};
#endif // CUTLASS_ARCH_MMA_SM103_SUPPORTED
template <typename Config>
typename Config::Gemm::Arguments args_from_options(
at::Tensor& D, at::Tensor const& A, at::Tensor const& B,
@@ -220,6 +362,38 @@ void cutlass_fp4_gemm_dispatch(torch::Tensor& D, torch::Tensor const& A,
}
}
// ============================================================================
// SM103 Dispatch
// ============================================================================
#if defined(CUTLASS_ARCH_MMA_SM103_SUPPORTED)
template <typename OutType>
void cutlass_fp4_gemm_sm103_dispatch(torch::Tensor& D, torch::Tensor const& A,
torch::Tensor const& B,
torch::Tensor const& A_sf,
torch::Tensor const& B_sf,
torch::Tensor const& alpha, int64_t m,
int64_t n, int64_t k,
cudaStream_t stream) {
uint32_t const mp2 = std::max(static_cast<uint32_t>(16), next_pow_2(m));
if (mp2 <= 16) {
// m in [1, 16] -- 1SM, low-latency decode
runGemm<Fp4GemmSm103<sm103_fp4_config_M16, OutType>>(
D, A, B, A_sf, B_sf, alpha, m, n, k, stream);
} else if (mp2 <= 256) {
// m in (16, 256] -- 2SM with moderate cluster
runGemm<Fp4GemmSm103<sm103_fp4_config_M256, OutType>>(
D, A, B, A_sf, B_sf, alpha, m, n, k, stream);
} else {
// m in (256, inf) -- 2SM with full cluster
runGemm<Fp4GemmSm103<sm103_fp4_config_default, OutType>>(
D, A, B, A_sf, B_sf, alpha, m, n, k, stream);
}
}
#endif // CUTLASS_ARCH_MMA_SM103_SUPPORTED
#else
template <typename OutType>
void cutlass_fp4_gemm_dispatch(torch::Tensor& D, torch::Tensor const& A,
@@ -315,3 +489,82 @@ void cutlass_scaled_fp4_mm_sm100a(torch::Tensor& D, torch::Tensor const& A,
")");
}
}
// ============================================================================
// SM103 Entry Point (B300 / Blackwell Ultra)
//
// Uses FP4 Ultra MMA instructions with K=768 tiles for higher throughput.
// Scale factors must be in Sm103BlockScaledConfig layout (different from SM100).
// ============================================================================
#if defined(CUTLASS_ARCH_MMA_SM103_SUPPORTED)
void cutlass_scaled_fp4_mm_sm103a(torch::Tensor& D, torch::Tensor const& A,
torch::Tensor const& B,
torch::Tensor const& A_sf,
torch::Tensor const& B_sf,
torch::Tensor const& alpha) {
CHECK_INPUT(A, FLOAT4_E2M1X2, "a");
CHECK_INPUT(B, FLOAT4_E2M1X2, "b");
CHECK_INPUT(A_sf, SF_DTYPE, "scale_a");
CHECK_INPUT(B_sf, SF_DTYPE, "scale_b");
CHECK_INPUT(alpha, at::ScalarType::Float, "alpha");
TORCH_CHECK(A.dim() == 2, "a must be a matrix");
TORCH_CHECK(B.dim() == 2, "b must be a matrix");
TORCH_CHECK(A.sizes()[1] == B.sizes()[1],
"a and b shapes cannot be multiplied (", A.sizes()[0], "x",
A.sizes()[1], " and ", B.sizes()[0], "x", B.sizes()[1], ")");
auto const m = A.sizes()[0];
auto const n = B.sizes()[0];
auto const k = A.sizes()[1] * 2;
constexpr int alignment = 32;
TORCH_CHECK(k % alignment == 0, "Expected k to be divisible by ", alignment,
", but got a shape: (", A.sizes()[0], "x", A.sizes()[1],
"), k: ", k, ".");
TORCH_CHECK(n % alignment == 0, "Expected n to be divisible by ", alignment,
", but got b shape: (", B.sizes()[0], "x", B.sizes()[1], ").");
// SM103 scale factor shape validation.
// Physical dimensions are the same as SM100 (padded to 128 x ceil(k/16,4)),
// but the internal swizzle pattern (Sm103BlockScaledConfig) differs.
auto round_up = [](int x, int y) { return (x + y - 1) / y * y; };
int rounded_m = round_up(m, 128);
int rounded_n = round_up(n, 128);
int rounded_k = round_up(k / 16, 4);
TORCH_CHECK(A_sf.dim() == 2, "scale_a must be a matrix");
TORCH_CHECK(B_sf.dim() == 2, "scale_b must be a matrix");
TORCH_CHECK(A_sf.sizes()[1] == B_sf.sizes()[1],
"scale_a and scale_b shapes cannot be multiplied (",
A_sf.sizes()[0], "x", A_sf.sizes()[1], " and ", B_sf.sizes()[0],
"x", B_sf.sizes()[1], ")");
TORCH_CHECK(A_sf.sizes()[0] == rounded_m && A_sf.sizes()[1] == rounded_k,
"scale_a must be padded and swizzled to a shape (", rounded_m,
"x", rounded_k, "), but got a shape (", A_sf.sizes()[0], "x",
A_sf.sizes()[1], ")");
TORCH_CHECK(B_sf.sizes()[0] == rounded_n && B_sf.sizes()[1] == rounded_k,
"scale_b must be padded and swizzled to a shape (", rounded_n,
"x", rounded_k, "), but got a shape (", B_sf.sizes()[0], "x",
B_sf.sizes()[1], ")");
auto out_dtype = D.dtype();
const at::cuda::OptionalCUDAGuard device_guard(device_of(A));
const cudaStream_t stream = at::cuda::getCurrentCUDAStream(A.get_device());
if (out_dtype == at::ScalarType::Half) {
cutlass_fp4_gemm_sm103_dispatch<cutlass::half_t>(D, A, B, A_sf, B_sf,
alpha, m, n, k, stream);
} else if (out_dtype == at::ScalarType::BFloat16) {
cutlass_fp4_gemm_sm103_dispatch<cutlass::bfloat16_t>(
D, A, B, A_sf, B_sf, alpha, m, n, k, stream);
} else {
TORCH_CHECK(false, "Unsupported output data type of nvfp4 mm (", out_dtype,
")");
}
}
#endif // CUTLASS_ARCH_MMA_SM103_SUPPORTED
+49
View File
@@ -199,6 +199,55 @@ __device__ __forceinline__ uint8_t* cvt_quant_to_fp4_get_sf_out_offset(
return reinterpret_cast<uint8_t*>(SFout) + SFOffset;
}
// ============================================================================
// SM103 (Blackwell Ultra / B300) swizzled SF offset.
//
// SM103 uses Sm103BlockScaledConfig with a 3-level M decomposition:
// M -> (m8, m4a, m4b) where mIdx = m4b + m4a*4 + m8*16
// K -> (sfv16_broadcast, k4)
//
// Atom layout:
// Shape: <Shape<_8, _4, _4>, Shape<SFVecSize=16, _4>>
// Stride: <Stride<_16, _128, _4>, Stride<_0, _1>>
//
// Physical offset = m8*16 + m4a*128 + m4b*4 + k4
// Each 128-row x 4-col tile occupies 512 bytes (same as SM100).
// ============================================================================
template <class SFType, int CVT_FP4_NUM_THREADS_PER_SF>
__device__ __forceinline__ uint8_t* cvt_quant_to_fp4_get_sf_out_offset_sm103(
int rowIdx, int colIdx, int32_t numKTiles, SFType* SFout) {
static_assert(CVT_FP4_NUM_THREADS_PER_SF == 1 ||
CVT_FP4_NUM_THREADS_PER_SF == 2);
if (threadIdx.x % CVT_FP4_NUM_THREADS_PER_SF != 0) {
return nullptr;
}
int32_t kIdx = colIdx / CVT_FP4_NUM_THREADS_PER_SF;
int32_t mIdx = rowIdx;
// SM103 tile decomposition (128 rows per M-tile, 4 K-positions per K-tile).
int32_t mTileIdx = mIdx >> 7; // mIdx / 128
int32_t mLocal = mIdx & 127; // mIdx % 128
// SM103 3-level M decomposition: mLocal = m4b + m4a*4 + m8*16
int32_t m4b = mLocal & 3; // mLocal % 4
int32_t m4a = (mLocal >> 2) & 3; // (mLocal / 4) % 4
int32_t m8 = (mLocal >> 4) & 7; // (mLocal / 16) % 8
int32_t kTileIdx = kIdx >> 2; // kIdx / 4
int32_t innerKIdx = kIdx & 3; // kIdx % 4
// Physical offset within the 512-byte tile:
// m8 * 16 + m4a * 128 + m4b * 4 + innerKIdx
// Tile base: (mTileIdx * numKTiles + kTileIdx) * 512
int64_t SFOffset = (static_cast<int64_t>(mTileIdx) * numKTiles + kTileIdx)
<< 9 |
(m8 << 4) | (m4a << 7) | (m4b << 2) | innerKIdx;
return reinterpret_cast<uint8_t*>(SFout) + SFOffset;
}
template <class SFType>
__device__ __forceinline__ uint8_t* sf_out_rowmajor_u8(int row, int pack,
int packs_per_row_sf,
@@ -107,12 +107,23 @@ def prepare_weights_for_nvfp4_flashinfer_trtllm(
def prepare_weights_for_nvfp4_cutlass(
weight: torch.Tensor,
weight_scale: torch.Tensor,
use_sm103_layout: bool = False,
) -> tuple[torch.Tensor, torch.Tensor, int]:
"""
Prepare weights and scales for CUTLASS/FlashInfer-CUTLASS FP4 GEMM.
This involves padding weights for alignment (K and N divisible by 32)
and swizzling scales to the layout expected by the target GPU.
Parameters
----------
use_sm103_layout : bool
If True, use the SM103 (B300) scale factor layout instead of SM100.
SM103 uses Sm103BlockScaledConfig with a 3-level M decomposition.
"""
swizzled_weight_scale = swizzle_blockscale(weight_scale)
if use_sm103_layout:
swizzled_weight_scale = swizzle_blockscale_sm103(weight_scale)
else:
swizzled_weight_scale = swizzle_blockscale(weight_scale)
padded_weight, weights_padding_cols = pad_nvfp4_weight_for_cutlass(weight)
return padded_weight, swizzled_weight_scale, weights_padding_cols
@@ -166,7 +177,8 @@ def convert_to_nvfp4_linear_kernel_format(
NvFp4LinearBackend.FLASHINFER_CUDNN,
):
weight, weight_scale, weights_padding_cols = prepare_weights_for_nvfp4_cutlass(
layer.weight.data, layer.weight_scale.data
layer.weight.data, layer.weight_scale.data,
use_sm103_layout=is_sm103(),
)
layer.weight = torch.nn.Parameter(weight, requires_grad=False)
layer.weight_scale = torch.nn.Parameter(weight_scale, requires_grad=False)
@@ -312,6 +324,95 @@ def swizzle_blockscale(scale: torch.Tensor) -> torch.Tensor:
return swizzled.reshape(B, M_padded, K_padded)
def swizzle_blockscale_sm103(scale: torch.Tensor) -> torch.Tensor:
"""
Pad and block-interleave the FP4 block-scales for SM103 (B300).
SM103 uses Sm103BlockScaledConfig with a different swizzle pattern:
M decomposition: m4b + m4a*4 + m8*16 (3-level, 4×4×8 = 128)
vs SM100:
M decomposition: outerM + innerM*32 (2-level, 32×4 = 128)
The K decomposition (4 per tile) is the same for both.
Parameters
----------
scale : torch.Tensor
FP8-E4M3FN block scales, shape [M, K/16] or [B, M, K/16].
Returns
-------
torch.Tensor
The SM103-swizzled tensor, same outer shape as *scale*.
"""
assert scale.dtype == torch.float8_e4m3fn, (
"swizzle_blockscale_sm103 expects the input tensor to be in "
"torch.float8_e4m3fn format."
)
scale_ndim = scale.ndim
if scale_ndim == 2:
scale = scale.unsqueeze(0)
assert scale.ndim == 3, "Expected a 2-D or 3-D tensor for block scales."
B, M, K = scale.shape
M_padded = round_up(M, 128)
K_padded = round_up(K, 4)
padded = torch.zeros(
(B, M_padded, K_padded), dtype=scale.dtype, device=scale.device
)
padded[:B, :M, :K] = scale
# SM103 3-level M decomposition: mLocal = m4b + m4a*4 + m8*16
# Reshape: [B, numMTiles, 8(m8), 4(m4a), 4(m4b), numKTiles, 4(innerK)]
padded = padded.reshape(
B, M_padded // 128, 8, 4, 4, K_padded // 4, 4
)
# Permute to: [B, mTile, kTile, m8, m4a, m4b, innerK]
# which matches stride layout: m8*16 + m4a*128 + m4b*4 + innerK
# In contiguous memory: last dims are innermost.
# Target byte offset = m8*16 + m4a*128 + m4b*4 + innerK
# We need the permutation that, when contiguous, produces these strides.
#
# Contiguous strides for shape [mTile, kTile, d0, d1, d2, d3]:
# d3 stride = 1 (innerK)
# d2 stride = 4 (m4b -> 4)
# d1 stride = 16 (m4a -> but we need 128!)
#
# Since contiguous layout assigns stride 1 to the last dim and
# increasing strides to earlier dims, we need to ORDER the dims
# so that the dim with stride 1 is last, stride 4 is second-to-last, etc.
#
# Target strides within 512-byte tile:
# innerK: stride 1
# m4b: stride 4
# m8: stride 16
# m4a: stride 128
#
# So ordering from outermost to innermost by decreasing stride:
# m4a (128) > m8 (16) > m4b (4) > innerK (1)
#
# Current dims: [B, mTile, m8, m4a, m4b, kTile, innerK]
# 0 1 2 3 4 5 6
# Target: [B, mTile, kTile, m4a, m8, m4b, innerK]
# 0 1 5 3 2 4 6
swizzled = padded.permute(0, 1, 5, 3, 2, 4, 6).contiguous().cuda()
if scale_ndim == 2:
return swizzled.reshape(M_padded, K_padded)
return swizzled.reshape(B, M_padded, K_padded)
def is_sm103() -> bool:
"""Check if the current device is SM103 (B300/Blackwell Ultra)."""
if not current_platform.is_cuda():
return False
cap = current_platform.get_device_capability()
return cap is not None and cap.to_int() == 103
def cutlass_fp4_supported() -> bool:
if not current_platform.is_cuda():
return False